// desktop.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"

#pragma pack (push, 1)

typedef struct
{
	DWORD  Data1;
	WORD   Data2;
	WORD   Data3;
	BYTE   Data4 [8];
} G;

typedef struct
{
	BYTE        bWidth;			// Width, in pixels, of the image
	BYTE        bHeight;		// Height, in pixels, of the image
	BYTE        bColorCount;	// Number of colors in image (0 if >=8bpp)
	BYTE        bReserved;		// Reserved ( must be 0)
	WORD        wPlanes;		// Color Planes
	WORD        wBitCount;		// Bits per pixel
	DWORD       dwBytesInRes;	// How many bytes in this resource?
	DWORD       dwImageOffset;	// Where in the file is this image?
} ICONDIRENTRY;

typedef struct
{ 
    BITMAPINFOHEADER bmiHeader;
    RGBQUAD bmiColors [256];
} BITMAPINFO256; 

typedef struct
{
	WORD idReserved;			// 0
	WORD idType;				// 1
	WORD idCount;				// 1
	ICONDIRENTRY idEntries;
	BITMAPINFO256 idBitmap;
} ICONDIR;

#pragma pack (pop)

const char* desktop_name =
	"Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\Streams\\Desktop";

const char* ini = ".\\desktop0.ini";

const struct
{
	const char *clsid;
	const char *command;
	int id;
} shell_object [] =
{
	{"{20D04FE0-3AEA-1069-A2D8-08002B30309D}", "mycomputer", CSIDL_DRIVES},	// My Computer
	{"{208D2C60-3AEA-1069-A2D7-08002B30309D}", "network", CSIDL_NETWORK},	// Network Neighborhood
	{"{645FF040-5081-101B-9F08-00AA002F954E}", "recycle", CSIDL_BITBUCKET},	// Recycle Bin
	{"{00020D76-0000-0000-C000-000000000046}", "inbox", CSIDL_DESKTOP},		// Inbox ( CSIDL_)
	{"{2227A280-3AEA-1069-A2DE-08002B30309D}", "printers", CSIDL_PRINTERS},	// Printers
	{"{450D8FBA-AD25-11D0-98A8-0800361B1103}", "mydoc", CSIDL_PERSONAL},	// My Documents
	{"{3DC7A020-0ACD-11CF-A9BB-00AA004AE837}", "", CSIDL_INTERNET},			// The Internet
	{"{871C5380-42A0-1069-A2EA-08002B30309D}", "", CSIDL_INTERNET},			// Internet Explorer
	{NULL, NULL, 0}
};

char object [20];
char ie_path [MAX_PATH];

void parse_desktop (const char* buf, int size);
void parse_clsid (int x, int y, const G* g);
void parse_link (int x, int y, const char* long_name, const char *short_name);
void save_file_icon (char *filename, int index, const char *iconname);
void save_icon (HICON icon, const char *iconname);
bool get_wallpaper (char *path);
COLORREF get_background ();
int get_wallpaper_style ();

int main(int argc, char* argv[])
{
	CoInitialize (NULL);

	//   
	CopyFile (ini, ".\\desktop0.bak", FALSE);

	//    Internet Explorer
	HKEY key = NULL;
	DWORD size = sizeof (ie_path);
	DWORD type = REG_SZ;
	if (RegOpenKeyEx (HKEY_LOCAL_MACHINE,
		"Software\\Microsoft\\IE Setup\\Setup", 0, KEY_READ, &key) == ERROR_SUCCESS &&
		RegQueryValueEx (key, "Path", NULL, &type, (LPBYTE) ie_path, &size) == ERROR_SUCCESS &&
		size > 1) {
		//    Internet Explorer
		strcat (ie_path, "\\iexplore.exe");
	}
	if (key) {
		RegCloseKey (key);
		key = NULL;
	}

	//  
	char wallpaper [MAX_PATH];
	get_wallpaper (wallpaper + 1);
	wallpaper [0] = '\"';
	strcat (wallpaper, "\"");
	strlwr (wallpaper);
	printf ("Wallpaper: \"%s\"\n", wallpaper);
	WritePrivateProfileString ("desktop0", "wallpaper", wallpaper, ini);

	//  
	char color [30];
	itoa (get_background (), color, 10);
	printf ("Color: %s\n", color);
	WritePrivateProfileString ("desktop0", "color", color, ini);

	//   
	char style [30];
	itoa (get_wallpaper_style (), style, 10);
	printf ("Style: %s\n", style);
	WritePrivateProfileString ("desktop0", "wallpaper style", style, ini);

	//  
	if (RegOpenKeyEx (HKEY_CURRENT_USER, desktop_name,
		0, KEY_READ, &key) == ERROR_SUCCESS) {
		type = REG_BINARY;
		size = 0;
		if (RegQueryValueEx (key, "ViewView2", NULL, &type, NULL, &size)
			== ERROR_SUCCESS && (type == REG_BINARY) && size >= 0xa0 + 26) {
			char *buf = new char [size];
			if (buf) {
				if (RegQueryValueEx (key, "ViewView2", NULL, &type, (LPBYTE) buf, &size)
					== ERROR_SUCCESS) {
					parse_desktop (buf, size);
				}
				delete buf;
			}
		}
		RegCloseKey (key);
	}

	//   
	HWND talisman = FindWindow ("TForm1", "Talisman");
	if (talisman)
		PostMessage (talisman, WM_USER + 2766, 0, 4);

	CoUninitialize ();

	return 0;
}

void parse_desktop (const char* buf, int size)
{
	int i = 1;
	for (const char *p = buf + buf [14]; p - buf < size - 24; i++) {
		sprintf (object, "button_%02d", i);	
		WritePrivateProfileString (object, "type", "1", ini);
		WritePrivateProfileString (object, "align", "1", ini);
		WritePrivateProfileString (object, "x label", "0", ini);
		WritePrivateProfileString (object, "y label", "0", ini);
		WritePrivateProfileString (object, "x icon", "5", ini);
		WritePrivateProfileString (object, "y icon", "-35", ini);
		WritePrivateProfileString (object, "width", "65", ini);
		WritePrivateProfileString (object, "height", "65", ini);

		int x = *((int*) p) - 5;
		p += 4;
		
		char sx [30];
		itoa (x, sx, 10);
		WritePrivateProfileString (object, "x", sx, ini);

		int y = *((int*) p);
		p += 4;
		
		char sy [30];
		itoa (y, sy, 10);
		WritePrivateProfileString (object, "y", sy, ini);

		DWORD r = *((DWORD*) p) & 0x00ff0000;
		p += 4;
		if (r == 0x001f0000) {
			// CLSID
			parse_clsid (x, y, (const G*) p);
			p += sizeof (G);			
		} else {
			// LINK
			p += 10;

			const char *long_name = p;
			p += strlen (long_name) + 1;

			const char *short_name = p;
			p += strlen (short_name) + 1;

			parse_link (x, y, long_name, short_name);
		}
		
		printf ("\n");
	}
}

void parse_clsid (int x, int y, const G* g)
{
	//  GUID   
	char clsid_name [60];
	sprintf (clsid_name, "{%08X-%04X-%04X-%02X%02X-%02X%02X%02X%02X%02X%02X}",
		g->Data1, g->Data2, g->Data3,
		g->Data4 [0], g->Data4 [1], g->Data4 [2], g->Data4 [3],
		g->Data4 [4], g->Data4 [5], g->Data4 [6], g->Data4 [7]);

	printf ("%s ", clsid_name);

	char name [MAX_PATH];
	strcpy (name, "Software\\Classes\\CLSID\\");
	strcat (name, clsid_name);

	char command [MAX_PATH];
	*command = '\0';
	for (int i = 0; shell_object [i].clsid; i++)
		if (!stricmp (shell_object [i].clsid, clsid_name)) {
			if (*(shell_object [i].command) == '\0') {
				strcpy (command, "%");
				strcat (command, ie_path);
				strcat (command, "\"");
			} else
				strcpy (command, shell_object [i].command);
			break;
		}
	WritePrivateProfileString (object, "command", command, ini);

	ITEMIDLIST* pidl = NULL;

	//  
	char title [MAX_PATH];
	DWORD size = sizeof (title);
	DWORD type = REG_SZ;
	HKEY key = NULL;
	if (RegOpenKeyEx (HKEY_CURRENT_USER, name, 0, KEY_READ, &key) == ERROR_SUCCESS &&
		RegQueryValueEx (key, "",NULL, &type, (LPBYTE) title, &size) == ERROR_SUCCESS &&
		size > 1) {
		//   HKEY_CURRENT_USER
	} else {
		size = sizeof (title);
		type = REG_SZ;
		if (key) {
			RegCloseKey (key);
			key = NULL;
		}
		if (RegOpenKeyEx (HKEY_LOCAL_MACHINE, name, 0, KEY_READ, &key) == ERROR_SUCCESS &&
			RegQueryValueEx (key, "",NULL, &type, (LPBYTE) title, &size) == ERROR_SUCCESS &&
			size > 1) {
			//   HKEY_LOCAL_MACHINE
		} else {
			SHFILEINFO sfi;
			memset (&sfi, 0, sizeof (sfi));
			SHGetSpecialFolderLocation (GetDesktopWindow (), shell_object [i].id, &pidl);
			SHGetFileInfo ((LPCTSTR) pidl, FILE_ATTRIBUTE_NORMAL, &sfi,
				sizeof (SHFILEINFO), SHGFI_PIDL | SHGFI_DISPLAYNAME);
			if (*(sfi.szDisplayName))
				//   -
				strcpy (title, sfi.szDisplayName);
			else
				//  ,   
				strcpy (title, command);
		}
	}
	if (key) {
		RegCloseKey (key);
		key = NULL;
	}

	if (pidl) {
		LPMALLOC pMalloc;
		SHGetMalloc (&pMalloc);
		pMalloc->Free (pidl);
		pMalloc->Release ();
	}

	printf ("%s ", title);
	WritePrivateProfileString (object, "text", title, ini);

	//  
	char path [MAX_PATH];
	int index = 0;
	size = sizeof (path);
	type = REG_SZ;
	strcat (name, "\\DefaultIcon");
	if (RegOpenKeyEx (HKEY_CURRENT_USER, name, 0, KEY_READ, &key) == ERROR_SUCCESS &&
		RegQueryValueEx (key, "", NULL, &type, (LPBYTE) path, &size) == ERROR_SUCCESS &&
		size > 1) {
		//   HKEY_CURRENT_USER
	} else {
		size = sizeof (path);
		type = REG_SZ;
		if (key) {
			RegCloseKey (key);
			key = NULL;
		}
		if (RegOpenKeyEx (HKEY_LOCAL_MACHINE, name, 0, KEY_READ, &key) == ERROR_SUCCESS &&
			RegQueryValueEx (key, "", NULL, &type, (LPBYTE) path, &size) == ERROR_SUCCESS &&
			size > 1) {
			//   HKEY_LOCAL_MACHINE
		} else
			//  
			if (!stricmp (clsid_name, "{871C5380-42A0-1069-A2EA-08002B30309D}"))
				//  Internet Explorer
				strcpy (path, ie_path);
			else {
				index = 3;
				strcpy (path, "shell32.dll");
			}
	}
	if (key) {
		RegCloseKey (key);
		key = NULL;
	}

	//  
	char *d = strrchr (path, ',');
	if (d) {
		*d = '\0';
		index = atoi (d + 1);
	}
	char iconname [MAX_PATH];
	strcpy (iconname, title);
	strcat (iconname, ".ico");
	save_file_icon (path, index, iconname);
	WritePrivateProfileString (object, "icon", iconname, ini);
}

void parse_link (int x, int y, const char* long_name, const char *short_name)
{
	printf ("\"%s\"", long_name);

	//   
	char path [MAX_PATH];
	SHGetSpecialFolderPath (NULL, path, CSIDL_DESKTOPDIRECTORY, FALSE);
	strcat (path, "\\");
	strcat (path, long_name);

	//  
	char command [MAX_PATH];
	strcpy (command, "%");
	strcat (command, path);
	strcat (command, "\"");
	WritePrivateProfileString (object, "command", command, ini);

	//  
	char title [MAX_PATH];
	strcpy (title, long_name);
	char *c = strrchr (title, '.');
	if (c && !stricmp (c, ".lnk"))
		*c = '\0';
	WritePrivateProfileString (object, "text", title, ini);

	//  
	char iconname [MAX_PATH];
	strcpy (iconname, title);
	strcat (iconname, ".ico");
	save_file_icon (path, 0, iconname);
	WritePrivateProfileString (object, "icon", iconname, ini);
}

void save_file_icon (char *filename, int index, const char *iconname)
{
	//      
	HICON icon = NULL;
	if (ExtractIconEx (filename, index, &icon, NULL, 1) != 1) {
		//    - 
		WORD icon_index = index;
		icon = ExtractAssociatedIcon (NULL, filename, &icon_index);
	}
	if (icon) {
		save_icon (icon, iconname);
		DestroyIcon (icon);
	}
}

void save_icon (HICON icon, const char *iconname)
{
	HDC hdc = GetDC (NULL);

	ICONINFO ii;
	GetIconInfo (icon, &ii);

	//  
	BITMAP image_bitmap;
	GetObject (ii.hbmColor, sizeof (BITMAP), &image_bitmap);

	BITMAPINFO256 image_info;
	memset (&image_info, 0, sizeof (image_info));
	image_info.bmiHeader.biSize = sizeof (BITMAPINFOHEADER);
	image_info.bmiHeader.biWidth = (BYTE) image_bitmap.bmWidth;
	image_info.bmiHeader.biHeight = (BYTE) image_bitmap.bmHeight;
	image_info.bmiHeader.biPlanes = 1;
	image_info.bmiHeader.biBitCount = 8;
	image_info.bmiHeader.biCompression = BI_RGB;

	int image_size = ((((image_info.bmiHeader.biWidth * image_info.bmiHeader.biBitCount) +
		31) & ~31) >> 3) * image_info.bmiHeader.biHeight;
	char *image_bits = new char [image_size];
	if (image_bits) {
		memset (image_bits, 0, image_size);

		GetDIBits (hdc, ii.hbmColor, 0, image_info.bmiHeader.biHeight,
			image_bits, (BITMAPINFO*) &image_info, DIB_RGB_COLORS);
	}

	//  
	BITMAP mask_bitmap;
	GetObject (ii.hbmMask, sizeof (BITMAP), &mask_bitmap);

	BITMAPINFO256 mask_info;
	memset (&mask_info, 0, sizeof (mask_info));
	mask_info.bmiHeader.biSize = sizeof (BITMAPINFOHEADER);
	mask_info.bmiHeader.biWidth = (BYTE) mask_bitmap.bmWidth;
	mask_info.bmiHeader.biHeight = (BYTE) mask_bitmap.bmHeight;
	mask_info.bmiHeader.biPlanes = 1;
	mask_info.bmiHeader.biBitCount = 1;
	mask_info.bmiHeader.biCompression = BI_RGB;

	int mask_size = ((((mask_info.bmiHeader.biWidth * mask_info.bmiHeader.biBitCount) +
		31) & ~31) >> 3) * mask_info.bmiHeader.biHeight;
	char *mask_bits = new char [mask_size];
	if (mask_bits) {
		memset (mask_bits, 0, mask_size);

		GetDIBits (hdc, ii.hbmMask, 0, mask_info.bmiHeader.biHeight,
			mask_bits, (BITMAPINFO*) &mask_info, DIB_RGB_COLORS);
	}

	if (image_bits && mask_bits) {
		//   
		FILE* file = fopen (iconname, "wb");
		if (file) {
			//   
			ICONDIR dir;
			memset (&dir, 0, sizeof (dir));
			dir.idType = 1;
			dir.idCount = 1;
			dir.idEntries.bWidth = (BYTE) image_info.bmiHeader.biWidth;
			dir.idEntries.bHeight = (BYTE) image_info.bmiHeader.biHeight;
			dir.idEntries.wPlanes = 1;
			dir.idEntries.wBitCount = 8;
			dir.idEntries.dwBytesInRes = sizeof (BITMAPINFO256) + image_size + mask_size;
			dir.idEntries.dwImageOffset = sizeof (ICONDIRENTRY) + 3 * sizeof (WORD);
			dir.idBitmap.bmiHeader.biSize = sizeof (BITMAPINFOHEADER);
			dir.idBitmap.bmiHeader.biWidth = image_info.bmiHeader.biWidth;
			dir.idBitmap.bmiHeader.biHeight = 2 * image_info.bmiHeader.biHeight;
			dir.idBitmap.bmiHeader.biPlanes = image_info.bmiHeader.biPlanes;
			dir.idBitmap.bmiHeader.biBitCount = image_info.bmiHeader.biBitCount;
			dir.idBitmap.bmiHeader.biSizeImage = image_size + mask_size;
			memcpy (dir.idBitmap.bmiColors, image_info.bmiColors, 256 * sizeof (RGBQUAD));

			//   
			fwrite (&dir, sizeof (ICONDIR), 1, file);

			//   
			fwrite (image_bits, image_size, 1, file);

			//  
			fwrite (mask_bits, mask_size, 1, file);

			fclose (file);
		}
	}

	if (image_bits)
		delete image_bits;

	if (mask_bits)
		delete mask_bits;

	ReleaseDC (NULL, hdc);
}

bool get_wallpaper (char *path)
{
	bool ret = false;
	*path = '\0';
	HKEY key;
	if (RegOpenKeyEx (HKEY_CURRENT_USER, "Control Panel\\Desktop",
		0, KEY_READ, &key) == ERROR_SUCCESS) {
		DWORD size = MAX_PATH;
		DWORD type = REG_SZ;
		if (RegQueryValueEx (key, "Wallpaper",
			NULL, &type, (LPBYTE) path, &size) == ERROR_SUCCESS) {
			ret = true;
		}
		RegCloseKey (key);
	}
	return ret;
}

int get_wallpaper_style ()
{
	int style = 0;
	HKEY key;
	if (RegOpenKeyEx (HKEY_CURRENT_USER, "Control Panel\\Desktop",
		0, KEY_READ, &key) == ERROR_SUCCESS) {
		char val [2];
		DWORD size = sizeof (val);
		DWORD type = REG_SZ;
		if (RegQueryValueEx (key, "TileWallpaper",
			NULL, &type, (LPBYTE) val, &size) == ERROR_SUCCESS) {
			if (*val == '1')
				// Tiled
				style = 1;
			else {
				size = sizeof (val);
				type = REG_SZ;
				if (RegQueryValueEx (key, "WallpaperStyle",
					NULL, &type, (LPBYTE) val, &size) == ERROR_SUCCESS) {
					if (*val == '2')
						// Stretched
						style = 2;
					else
						// Centered
						style = 0;
				}
			}
		}
		RegCloseKey (key);
	}
	return style;
}

COLORREF get_background ()
{
	COLORREF color = 0;
	HKEY key;
	if (RegOpenKeyEx (HKEY_CURRENT_USER, "Control Panel\\Colors",
		0, KEY_READ, &key) == ERROR_SUCCESS) {
		char color_name [15];
		DWORD size = sizeof (color_name);
		DWORD type = REG_SZ;
		if (RegQueryValueEx (key, "Background",
			NULL, &type, (LPBYTE) color_name, &size) == ERROR_SUCCESS) {
			int red, green, blue;
			if (sscanf (color_name, " %d %d %d ", &red, &green, &blue) == 3) {
				color = RGB (red, green, blue);
			}
		}
		RegCloseKey (key);
	}
	return color;
}
